#######################################################################
# Visitor translation class
#######################################################################


from __future__ import unicode_literals
from __future__ import print_function

import re

from arpeggio import PTNodeVisitor
from pymuv.errors import MuvError
from pymuv.lvalue import LValue
from pymuv.settable import Settable
import pymuv.muvnodes as mn


class MuvVisitor(PTNodeVisitor):
    def __init__(self, *args, **kwargs):
        super(MuvVisitor, self).__init__(*args, **kwargs)

    def indent(self, txt, ind='    '):
        return "\n".join((ind + l if l else l) for l in txt.split('\n'))

    def left_associate_tree(self, arr):
        out = arr[0]
        arr = arr[1:]
        while len(arr) > 1:
            oper, expr = arr[:2]
            arr = arr[2:]
            out = mn.MuvNodeBinaryExpr(expr.position, out, oper, expr)
        return out

    def visit_ows(self, node, children):
        return None

    def visit_ws(self, node, children):
        return None

    def visit_SEMICOLON(self, node, children):
        return None

    def visit_kwend(self, node, children):
        return None

    def visit_numsign(self, node, children):
        if str(children[0]) == "-":
            return "-"
        return None

    def visit_bin_integer(self, node, children):
        chars = children[0].replace('_', '')
        val = 0
        for ch in chars:
            val *= 2
            val += int(ch)
        return mn.MuvNodeInteger(node.position, val)

    def visit_oct_integer(self, node, children):
        chars = children[0].replace('_', '')
        val = 0
        for ch in chars:
            val *= 8
            val += int(ch)
        return mn.MuvNodeInteger(node.position, val)

    def visit_dec_integer(self, node, children):
        chars = children[-1].replace('_', '')
        val = 0
        for ch in chars:
            val *= 10
            val += int(ch)
        return mn.MuvNodeInteger(node.position, val)

    def visit_hex_integer(self, node, children):
        chars = children[0].replace('_', '')
        val = 0
        for ch in chars:
            val *= 16
            val += "0123456789abcdef".index(str(ch).lower())
        return mn.MuvNodeInteger(node.position, val)

    def visit_unsigned_integer(self, node, children):
        return children[0]

    def visit_intnum(self, node, children):
        val = children[-1]
        if children[0] == "-":
            val.value *= -1
        return val

    def visit_raw_dquote_string(self, node, children):
        return children[0][2:-1]

    def visit_raw_squote_string(self, node, children):
        return children[0][2:-1]

    def visit_raw_dquote3_string(self, node, children):
        return children[0][4:-3]

    def visit_raw_squote3_string(self, node, children):
        return children[0][4:-3]

    def visit_dquote_string(self, node, children):
        return children[0][1:-1]

    def visit_squote_string(self, node, children):
        return children[0][1:-1]

    def visit_dquote3_string(self, node, children):
        return children[0][3:-3]

    def visit_squote3_string(self, node, children):
        return children[0][3:-3]

    def decode_string(self, txt):
        out = ''
        while txt:
            if '\\' not in txt:
                out += txt
                break
            pre, txt = txt.split('\\', 1)
            out += pre
            ch = txt[0]
            txt = txt[1:]
            if ch == '\n':
                continue
            elif ch == 'r' or ch == 'n':
                ch = '\r'
            elif ch == 'e' or ch == '[':
                ch = '\033'
            out += ch
        return out

    def visit_raw_string_literal(self, node, children):
        return children[0]

    def visit_esc_string_literal(self, node, children):
        return self.decode_string(children[0])

    def visit_string_literal(self, node, children):
        return mn.MuvNodeString(node.position, children[0])

    def visit_dbref(self, node, children):
        return mn.MuvNodeDbRef(node.position, children[0].value)

    def visit_floatnum(self, node, children):
        res = float("".join(children))
        return mn.MuvNodeFloat(node.position, res)

    def visit_OPER_INCR(self, node, children):
        return "++"

    def visit_OPER_DECR(self, node, children):
        return "--"

    def visit_OPER_PLUS(self, node, children):
        return "+"

    def visit_OPER_MINUS(self, node, children):
        return "-"

    def visit_OPER_MULT(self, node, children):
        return "*"

    def visit_OPER_DIV(self, node, children):
        return "/"

    def visit_OPER_MOD(self, node, children):
        return "%"

    def visit_OPER_POWER(self, node, children):
        return "pow"

    def visit_OPER_BITLEFT(self, node, children):
        return "bitshift"

    def visit_OPER_BITRIGHT(self, node, children):
        return "-1 * bitshift"

    def visit_OPER_BITAND(self, node, children):
        return "bitand"

    def visit_OPER_BITOR(self, node, children):
        return "bitor"

    def visit_OPER_BITXOR(self, node, children):
        return "bitxor"

    def visit_OPER_BITNOT(self, node, children):
        return "-1 bitxor"

    def visit_ASGN_PLUS(self, node, children):
        return "+"

    def visit_ASGN_MINUS(self, node, children):
        return "-"

    def visit_ASGN_MULT(self, node, children):
        return "*"

    def visit_ASGN_DIV(self, node, children):
        return "/"

    def visit_ASGN_MOD(self, node, children):
        return "%"

    def visit_ASGN_POWER(self, node, children):
        return "pow"

    def visit_ASGN_BITLEFT(self, node, children):
        return "bitshift"

    def visit_ASGN_BITRIGHT(self, node, children):
        return "-1 * bitshift"

    def visit_ASGN_BITAND(self, node, children):
        return "bitand"

    def visit_ASGN_BITOR(self, node, children):
        return "bitor"

    def visit_ASGN_BITXOR(self, node, children):
        return "bitxor"

    def visit_COMP_LT(self, node, children):
        return "<"

    def visit_COMP_GT(self, node, children):
        return ">"

    def visit_COMP_LTE(self, node, children):
        return "<="

    def visit_COMP_GTE(self, node, children):
        return ">="

    def visit_COMP_EQ(self, node, children):
        return "="

    def visit_COMP_NEQ(self, node, children):
        return "= not"

    def visit_COMP_STREQ(self, node, children):
        return "strcmp not"

    def visit_COMP_IN(self, node, children):
        return "swap array_findval"

    def visit_LOGICAL_AND(self, node, children):
        return None

    def visit_LOGICAL_OR(self, node, children):
        return None

    def visit_LOGICAL_XOR(self, node, children):
        return "xor"

    def visit_LOGICAL_NOT(self, node, children):
        return "not"

    def visit_paren_expr(self, node, children):
        return children[0]

    def visit_tuple_parts(self, node, children):
        return Settable(*children)

    def visit_settable(self, node, children):
        return children[0]

    def visit_declared_lvalue(self, node, children):
        varname = str(children[0])
        return LValue(varname, [], node.position, declare=True)

    def visit_indexed_lvalue(self, node, children):
        varname = str(children[0])
        indexing = children[1:]
        return LValue(varname, indexing, node.position)

    def visit_lvalue(self, node, children):
        return children[0]

    def visit_leaf_variable(self, node, children):
        varname = "".join(children)
        return mn.MuvNodeVarFetch(
            node.position,
            LValue(varname, [], node.position)
        )

    def visit_variable(self, node, children):
        return children[0]

    def visit_for_expr(self, node, children):
        stride = mn.MuvNodeInteger(node.position, 1)
        if len(children) > 3:
            lvalue, start, end, stride = children
        else:
            lvalue, start, end = children
        return mn.MuvNodeForLoop(
            node.position,
            var=lvalue,
            start=start,
            end=end,
            stride=stride,
        )

    def visit_foreach_expr(self, node, children):
        if len(children) > 2:
            keysettable, valsettable, expr = children
        else:
            valsettable, expr = children
            keysettable = None
        return mn.MuvNodeForEachLoop(
            node.position,
            expr=expr,
            keyvar=keysettable,
            valvar=valsettable,
        )

    def visit_compr_for(self, node, children):
        return children[0]

    def visit_compr_cond_if(self, node, children):
        return mn.MuvNodeIfElse(children[0].position, cond=children[0])

    def visit_compr_cond_unless(self, node, children):
        return mn.MuvNodeIfElse(
            children[0].position,
            cond=mn.MuvNodeCommand(children[0].position, 'not', children[0])
        )

    def visit_compr_cond(self, node, children):
        return children[0]

    def visit_keyval_pair(self, node, children):
        return mn.MuvNodeKeyValPair(node.position, children[0], children[1])

    def visit_dictlist(self, node, children):
        return children

    def visit_dict_initialization(self, node, children):
        vals = []
        if children:
            vals = children[0]
            if vals == ['=>']:
                vals = []
        return mn.MuvNodeWrappedExpr(
            node.position,
            "{", "}dict",
            *vals
        )

    def visit_list_initialization(self, node, children):
        vals = []
        if children:
            vals = children[0]
        return mn.MuvNodeWrappedExpr(
            node.position,
            "{", "}list",
            *vals
        )

    def visit_dict_comprehension(self, node, children):
        cond = None
        if len(children) > 2:
            loop, cond, body = children
        else:
            loop, body = children
        body = mn.MuvNodeCommand(body.position, "rot rot ->[]", body)
        if cond:
            cond.ifclause = body
            body = cond
        loop.body = body
        return mn.MuvNodeExprList(node.position, "{ }dict", loop)

    def visit_list_comprehension(self, node, children):
        cond = None
        if len(children) > 2:
            loop, cond, body = children
        else:
            loop, body = children
        body = mn.MuvNodeCommand(body.position, "swap []<-", body)
        if cond:
            cond.ifclause = body
            body = cond
        loop.body = body
        return mn.MuvNodeExprList(node.position, "{ }list", loop)

    def visit_comprehension(self, node, children):
        return children[0]

    def visit_leaf_expr(self, node, children):
        return children[0]

    def visit_attribute(self, node, children):
        attr = str(children[0])
        return mn.MuvNodeString(node.position, attr)

    def visit_index_part(self, node, children):
        return children[0]

    def visit_index_oper(self, node, children):
        idx = children[0]
        out = mn.MuvNodeIndex(node.position, idx=idx)
        return out

    def visit_addrcall_oper(self, node, children):
        args = children[0]
        out = mn.MuvNodeAddrCall(node.position, args=args)
        return out

    def visit_subscript_expr(self, node, children):
        out = children[0]
        for oper in children[1:]:
            oper.expr = out
            out = oper
        return out

    def visit_postfix_expr(self, node, children):
        if len(children) == 1:
            return children[0]
        return mn.MuvNodePostfixExpr(
            node.position,
            children[0],
            children[1],
        )

    def visit_unary_expr(self, node, children):
        if len(children) == 1:
            return children[0]
        oper = str(children[0])
        if oper == "+":
            return children[1]
        if oper == "-":
            return mn.MuvNodeBinaryExpr(
                node.position,
                children[1], "*", mn.MuvNodeInteger(node.position, -1)
            )
        if oper in ["++", "--"]:
            return mn.MuvNodePrefixExpr(node.position, oper, children[1])
        return mn.MuvNodeExprList(node.position, children[1], oper)

    def visit_power_expr(self, node, children):
        return self.left_associate_tree(children)

    def visit_multiplicative_expr(self, node, children):
        return self.left_associate_tree(children)

    def visit_additive_expr(self, node, children):
        return self.left_associate_tree(children)

    def visit_shift_expr(self, node, children):
        return self.left_associate_tree(children)

    def visit_relational_expr(self, node, children):
        return self.left_associate_tree(children)

    def visit_equality_expr(self, node, children):
        return self.left_associate_tree(children)

    def visit_bitand_expr(self, node, children):
        return self.left_associate_tree(children)

    def visit_bitxor_expr(self, node, children):
        return self.left_associate_tree(children)

    def visit_bitor_expr(self, node, children):
        return self.left_associate_tree(children)

    def visit_logical_and_expr(self, node, children):
        out = children[-1]
        for expr in reversed(children[:-1]):
            out = mn.MuvNodeLogicalAnd(expr.position, expr, out)
        return out

    def visit_logical_xor_expr(self, node, children):
        return self.left_associate_tree(children)

    def visit_logical_or_expr(self, node, children):
        out = children[-1]
        for expr in reversed(children[:-1]):
            pos = node.position
            if hasattr(expr, 'position'):
                pos = expr.position
            out = mn.MuvNodeLogicalOr(pos, expr, out)
        return out

    def visit_conditional_expr(self, node, children):
        if len(children) == 1:
            return children[0]
        return mn.MuvNodeIfElse(
            node.position,
            cond=children[0],
            ifclause=children[1],
            elseclause=children[2],
        )

    def visit_asgn_expr(self, node, children):
        lvalue, oper, value = children[:3]
        return mn.MuvNodeVarAssign(node.position, lvalue, value)

    def visit_asgn_append_expr(self, node, children):
        lvalue, oper, value = children[:3]
        return mn.MuvNodeVarOperAssign(
            node.position, lvalue, "swap []<-", value
        )

    def visit_asgn_oper_expr(self, node, children):
        lvalue, oper, value = children[:3]
        return mn.MuvNodeVarOperAssign(node.position, lvalue, oper, value)

    def visit_assignment_expr(self, node, children):
        return children[0]

    def visit_expr(self, node, children):
        if len(children) == 1:
            return children[0]
        out = mn.MuvNodeStatements(node.position)
        for child in children[:-1]:
            out.append(mn.MuvNodeCommand(child.position, "pop", child))
        out.append(children[-1])
        return out

    def visit_stmt_return(self, node, children):
        if len(children):
            return mn.MuvNodeCommand(node.position, "exit", children[0])
        return mn.MuvNodeCommand(node.position, "exit")

    def visit_stmt_break(self, node, children):
        return mn.MuvNodeStatements(node.position, "break")

    def visit_stmt_continue(self, node, children):
        return mn.MuvNodeStatements(node.position, "continue")

    def visit_stmt_expr(self, node, children):
        expr = children[0]
        if type(expr) not in [mn.MuvNodeVarAssign, mn.MuvNodeVarOperAssign]:
            expr = mn.MuvNodeExprList(expr.position, expr, "pop")
        return expr

    def visit_simple_statement(self, node, children):
        return children[0]

    def visit_stmt_do_if(self, node, children):
        return mn.MuvNodeIfElse(
            node.position,
            cond=children[1],
            ifclause=children[0],
        )

    def visit_stmt_do_unless(self, node, children):
        return mn.MuvNodeIfElse(
            node.position,
            cond=mn.MuvNodeCommand(children[1].position, "not", children[1]),
            ifclause=children[0],
        )

    def visit_stmt_simple(self, node, children):
        return children[0]

    def visit_statements(self, node, children):
        if not children:
            return mn.MuvNodeNull(node.position)
        return mn.MuvNodeStatements(node.position, *children)

    def visit_statement(self, node, children):
        if not children:
            return mn.MuvNodeNull(node.position)
        return children[0]

    def visit_stmt_for(self, node, children):
        forloop, body = children
        forloop.body = body
        return forloop

    def visit_stmt_foreach(self, node, children):
        forloop, body = children
        forloop.body = body
        return forloop

    def visit_stmt_if_else(self, node, children):
        if len(children) > 2:
            return mn.MuvNodeIfElse(
                node.position,
                cond=children[0],
                ifclause=children[1],
                elseclause=children[2],
            )
        return mn.MuvNodeIfElse(
            node.position,
            cond=children[0],
            ifclause=children[1],
        )

    def visit_stmt_while(self, node, children):
        return mn.MuvNodeWhile(
            node.position,
            cond=children[0],
            body=children[1],
        )

    def visit_stmt_until(self, node, children):
        return mn.MuvNodeUntil(
            node.position,
            cond=children[0],
            body=children[1],
        )

    def visit_stmt_do_while(self, node, children):
        return mn.MuvNodeDoWhile(
            node.position,
            body=children[0],
            cond=children[1],
        )

    def visit_stmt_do_until(self, node, children):
        return mn.MuvNodeDoUntil(
            node.position,
            body=children[0],
            cond=children[1],
        )

    def visit_stmt_try(self, node, children):
        settable = None
        if len(children) > 2:
            body, settable, handler = children
        else:
            body, handler = children
        return mn.MuvNodeTryCatch(
            node.position,
            body=body,
            handler=handler,
            var=settable,
        )

    def visit_usable_comparator(self, node, children):
        return children[0]

    def visit_using_comparator(self, node, children):
        comp = children[0]
        if str(comp) in ["strcmp", "stringcmp"]:
            return mn.MuvNodeExprList(node.position, comp, "not")
        return mn.MuvNodeExprList(node.position, comp)

    def visit_using_raw_muf(self, node, children):
        return mn.MuvNodeStatements(node.position, children[0])

    def visit_using_function(self, node, children):
        return mn.MuvNodeStatements(node.position, children[0])

    def visit_expr_using(self, node, children):
        if len(children) == 1:
            children.append(mn.MuvNodeExprList(node.position, "="))
        return children

    def visit_case_clause(self, node, children):
        return children

    def visit_case_clauses(self, node, children):
        return children

    def visit_default_clause(self, node, children):
        return children[0]

    def visit_switch_default(self, node, children):
        if children:
            return mn.MuvNodeExprList(
                node.position,
                children[0],
                "break",
            )
        return None

    def visit_stmt_switch(self, node, children):
        if len(children) == 2:
            children.append('')
        (expr, using_clause), case_clauses, dflt_clause = children
        if type(using_clause) == str:
            using_clause = mn.MuvNodeString(node.position, using_clause)
        cases = []
        vname = "swvar"
        for val, stmt in case_clauses:
            cases.append(
                mn.MuvNodeIfElse(
                    val.position,
                    cond=mn.MuvNodeExprList(
                        val.position,
                        mn.MuvNodeVarFetch(
                            val.position,
                            LValue(vname, [], val.position)
                        ),
                        val,
                        using_clause,
                    ),
                    ifclause=mn.MuvNodeExprList(
                        stmt.position, stmt, "break"
                    )
                )
            )
        if dflt_clause:
            cases.append(dflt_clause)
        lvalue = LValue(vname, [], expr.position, declare=True)
        return mn.MuvNodeSwitch(
            node.position,
            expr=expr,
            var=lvalue,
            cases=cases,
        )

    def visit_gstmt_include(self, node, children):
        filename = children[0].value
        muvparser = self.muvparser
        if muvparser.debug:
            line = muvparser.get_parse_line(children[0].position)
            muvparser.output += '\n(MUV:L%d) ' % line
        muvparser.include_file(filename, children[0].position)
        return None

    def visit_KW_TOP(self, node, children):
        return mn.MuvNodeStatements(node.position, '(top)')

    def visit_del_expr(self, node, children):
        lvalue = children[0]
        return mn.MuvNodeVarDel(node.position, lvalue)

    def visit_push_expr(self, node, children):
        return mn.MuvNodeExprList(node.position, *children)

    def visit_fmtstring_expr(self, node, children):
        args = children[0]
        fmt = args[0]
        if isinstance(fmt, mn.MuvNodeString):
            pos = fmt.position
            fmt = fmt.value
            pats = re.findall(r'%.', fmt.replace('%%', ''))
            if len(args) - 1 != len(pats):
                raise MuvError(
                    (
                        "fmtstring(fmt, ...) format string "
                        "expects {exp} args, but got {found}."
                    ).format(
                        exp=len(pats),
                        found=len(args) - 1,
                    ),
                    position=pos,
                )
            errfmt = (
                "fmtstring(fmt, ...) format expected "
                "{exp} arg, but got {found} ({num})."
            )
            literals = [
                mn.MuvNodeInteger,
                mn.MuvNodeFloat,
                mn.MuvNodeDbRef,
                mn.MuvNodeString,
            ]
            fmt_types = {
                "%i": ("integer", mn.MuvNodeInteger),
                "%d": ("dbref", mn.MuvNodeDbRef),
                "%D": ("dbref", mn.MuvNodeDbRef),
                "%e": ("float", mn.MuvNodeFloat),
                "%f": ("float", mn.MuvNodeFloat),
                "%g": ("float", mn.MuvNodeFloat),
                "%s": ("string", mn.MuvNodeString),
            }
            num = 0
            for pat, val in zip(pats, args[1:]):
                num += 1
                if type(val) in literals:
                    typname, typ = fmt_types[pat]
                    if not isinstance(val, typ):
                        raise MuvError(
                            errfmt.format(
                                exp=typname,
                                found=val.valtype,
                                num=num,
                            ),
                            position=val.position,
                        )
        args = list(reversed(list(args)))
        return mn.MuvNodeCommand(node.position, "fmtstring", *args)

    def visit_raw_muf(self, node, children):
        raw = children[0].value
        return mn.MuvNodeStatements(node.position, raw)

    def visit_stmt_raw_muf(self, node, children):
        return children[0]

    def visit_gstmt_const(self, node, children):
        cname, oper, expr = children
        return mn.MuvNodeGlobalConst(
            node.position,
            cname, expr
        )
        return None

    def visit_stmt_const(self, node, children):
        cname, oper, expr = children
        return mn.MuvNodeFuncConst(
            node.position,
            cname, expr
        )

    def visit_stmt_var(self, node, children):
        vname = children[0]
        return mn.MuvNodeFuncVar(node.position, vname)

    def visit_arglist(self, node, children):
        return children

    def visit_NS_SEP(self, node, children):
        return '::'

    def visit_ns_parts(self, node, children):
        return "".join(children)

    def visit_NS_IDENT(self, node, children):
        return "".join(children)

    def visit_VAR_IDENT(self, node, children):
        return "".join(children)

    def visit_CONST_IDENT(self, node, children):
        return "".join(children)

    def visit_FUNC_IDENT(self, node, children):
        return "".join(children)

    def visit_function_addr(self, node, children):
        funcname = children[1]
        return mn.MuvNodeFuncAddr(node.position, funcname)

    def visit_function_call(self, node, children):
        if len(children) < 2:
            children.append([])
        funcname, argslist = children[:2]
        return mn.MuvNodeFuncCall(
            node.position, funcname, argslist
        )

    def visit_dir_language(self, node, children):
        if children[0].value != "muv":
            raise MuvError(
                'Only $language "muv" allowed.',
                position=node.position,
            )
        return None

    def visit_dir_version(self, node, children):
        return mn.MuvNodeStatements(
            node.position,
            "$version %.3f\n" % children[0].value
        )

    def visit_dir_libversion(self, node, children):
        return mn.MuvNodeStatements(
            node.position,
            "$lib-version %.3f\n" % children[0].value
        )

    def visit_dir_author(self, node, children):
        return mn.MuvNodeStatements(
            node.position,
            "$author %s\n" % children[0].value
        )

    def visit_dir_note(self, node, children):
        return mn.MuvNodeStatements(
            node.position,
            "$note %s\n" % children[0].value
        )

    def visit_dir_echo(self, node, children):
        return mn.MuvNodeStatements(
            node.position,
            "$echo %s\n" % children[0].value
        )

    def visit_dir_pragma(self, node, children):
        return mn.MuvNodeStatements(
            node.position,
            "$pragma %s\n" % children[0].value
        )

    def visit_dir_include(self, node, children):
        return mn.MuvNodeStatements(
            node.position,
            "$include %s\n" % children[0].value
        )

    def visit_dir_error(self, node, children):
        raise MuvError(
            children[0].value,
            position=node.position,
        )
        return None

    def visit_dir_warn(self, node, children):
        print(children[0])
        return None

    def visit_KW_VOID(self, node, children):
        return 0

    def visit_KW_SINGLE(self, node, children):
        return 1

    def visit_KW_MULTIPLE(self, node, children):
        return 99

    def visit_gstmt_var_declinit(self, node, children):
        return mn.MuvNodeGlobalVar(node.position, children[0], children[2])

    def visit_gstmt_var_declonly(self, node, children):
        return mn.MuvNodeGlobalVar(node.position, children[0])

    def visit_simple_extern(self, node, children):
        extern_type, extern_name, (argvars, varargs) = children
        argcnt = len(argvars)
        return mn.MuvNodeExtern(
            node.position,
            extern_name,
            retcount=extern_type,
            argcount=argcnt,
            varargs=varargs,
        )

    def visit_raw_muf_extern(self, node, children):
        extern_type, extern_name, (argvars, varargs), oper, code = children
        argcnt = len(argvars)
        return mn.MuvNodeExtern(
            node.position,
            extern_name,
            retcount=extern_type,
            argcount=argcnt,
            varargs=varargs,
            code=code.value,
        )

    def visit_KW_PUBLIC(self, node, children):
        return "public"

    def visit_opt_public(self, node, children):
        return children

    def visit_declist(self, node, children):
        return children

    def visit_VARARG_MARKER(self, node, children):
        return "..."

    def visit_argvar_list(self, node, children):
        if len(children) == 0:
            return ([], False)
        if len(children) == 1:
            return (children[0], False)
        return (children[0], True)

    def visit_gstmt_func(self, node, children):
        pub = False
        if len(children) > 3:
            children = children[1:]
            pub = True
        funcname, (arglist, varargs), body = children
        return mn.MuvNodeFuncDef(
            node.position,
            funcname=funcname,
            args=arglist,
            body=body,
            varargs=varargs,
            public=pub,
        )

    def visit_gstmt_using_ns(self, node, children):
        ns = children[0]
        return mn.MuvNodeUsingNamespace(node.position, ns)

    def visit_nsdecl(self, node, children):
        namespace, body = children
        return mn.MuvNodeNamespace(node.position, namespace, body)

    def visit_global_statements(self, node, children):
        return children

    def visit_source_file(self, node, children):
        children = children[0]
        return mn.MuvNodeSourceFile(node.position, *children)

    def visit_program(self, node, children):
        return mn.MuvNodeProgram(node.position, children[0])


# vim: set ts=4 sw=4 et ai hlsearch nowrap :



from __future__ import unicode_literals
from __future__ import print_function

import os
import sys
from pkg_resources import resource_string

from arpeggio import (
    ParserPython, visit_parse_tree, NoMatch,
)

from pymuv.grammar import program, source_file, comment
from pymuv.errors import MuvError
from pymuv.context import Context
from pymuv.visitor import MuvVisitor


class MuvParser(object):
    def __init__(self):
        self.input_sources = ''
        self.output = ''
        self.error_cb = None
        self.error_found = False
        self.context = Context()
        self.debug = False
        self.optimization_level = False
        self.sysincludes_only = False
        self.wrapper_program = None

    def set_debug(self, val=True):
        self.debug = val
        self.context.debug = val

    def print_error(self, filename, line, col, msg):
        self.error_found = True
        srclines = self.input_sources.split("\n")
        if filename:
            filename += " "
        else:
            filename = ""
        err = "Error in %sline %d, col %d: %s" % (
            filename, line, col, msg
        )
        if self.error_cb:
            self.error_cb(err)
        else:
            print(srclines[line-1], file=sys.stderr)
            print("%s^" % (' '*(col-1)), file=sys.stderr)
            print(err, file=sys.stderr)

    def simplify_parse_error(self, e):
        groups = {
            "end-of-file": ["EOF"],
            "string-delimiter": [
                '"', "'", '"""', "'''",
                'r"', "r'", 'r"""', "r'''",
            ],
            "identifier": [r'[a-zA-Z_][a-zA-Z_0-9]*'],
            "variable-declaration": ['var', 'const'],
            "global-statement": [
                'func', 'public', 'extern', 'namespace',
                'using', 'include'
            ],
            "$directive": [
                "$author", "$echo", "$error", "$include",
                "$language", "$libversion", "$note",
                "$pragma", "$version", "$warn",
            ],
            "comparator": [
                '==', '!(?!=)', '<', '>', '<=', '>=', 'in', 'eq'
            ],
            "expression": [
                '%(?!=)', '&(?![&=])', '(', ')', '-(?![=-])',
                '/(?![/*=])', '<(?![=<])', '<<(?!=)', '>(?![=>])',
                '>>(?!=)', '\*(?![/*=])', '\*\*(?!=)', '\+(?![=+])',
                '\|(?![|=])', '^(?![^=])',
                '+', '-', '*', '/', '%', '**',
                '<<', '>>', '&', '|', '^', '~',
                '++', '--', '.', '[', '?',
                '&&', '||', '^^', '!',
                '0b', '0o', '0d', '0x', '#',
                'top', 'push', 'del', 'fmtstring', 'muf',
                r'[01_]+', r'[0-7_]+', r'[0-9a-fA-F_]+',
                r'[0-9_]+(?![.eE])', r'[+-]?(\d+\.\d*|\d*\.\d+)',
                r'[+-]?(\d+\.\d*|\d*\.\d+|\d)[eE][+-]?[0-9]+',
            ],
            "assignment": [
                '=(?![=>])',
                '=', '+=', '-=', '*=', '/=', '%=',
                '**=', '<<=', '>>=', '&=', '|=', '^=',
            ],
        }
        unique_rules = {}
        for rule in e.rules:
            if rule.rule_name == "EOF":
                expect = "EOF"
            else:
                expect = rule.to_match
            found = False
            for group, pats in groups.items():
                if expect in pats:
                    expect = group
                    found = True
            if not found:
                expect = "'%s'" % expect
            unique_rules[expect] = 1
        expected = " or ".join(sorted(list(unique_rules.keys())))
        return expected

    def get_parse_line(self, pos):
        line, col = self.context.parsers[-1].pos_to_linecol(pos)
        return line

    def parse_string(self, src, grammar=program, filename=None):
        oldsrcs = self.input_sources
        self.context.optimization_level = self.optimization_level
        self.input_sources = src
        parser = ParserPython(
            grammar,
            comment_def=comment,
            skipws=True,
            reduce_tree=False,
            memoization=True,
            debug=False,
        )
        self.context.parsers.append(parser)
        self.context.filenames.append(filename)
        try:
            parse_tree = parser.parse(self.input_sources)
            visitor = MuvVisitor(debug=False)
            visitor.muvparser = self
            parse_tree = visit_parse_tree(parse_tree, visitor)
            out = parse_tree.generate_code(self.context)
            if self.error_found:
                return False
            if len(self.context.filenames) == 1:
                if self.context.filenames[-1]:
                    filetext = " from {0}".format(
                        self.context.filenames[-1]
                    )
                else:
                    filetext = ''
                self.output = (
                    "( Generated{0} by the MUV compiler. )\n"
                    "(   https://github.com/revarbat/pymuv )\n"
                    "{1}\n"
                ).format(filetext, self.output)
            self.output += out
            if not self.error_found and len(self.context.filenames) == 1:
                if self.wrapper_program:
                    self.output = (
                        "@program {0}\n"
                        "1 99999 d\n"
                        "1 i\n"
                        "{1}\n"
                        ".\n"
                        "c\n"
                        "q\n"
                    ).format(self.wrapper_program, self.output)
            return True
        except MuvError as e:
            line, col = parser.pos_to_linecol(e.position)
            self.print_error(filename, line, col, str(e))
            return False
        except NoMatch as e:
            line, col = parser.pos_to_linecol(e.position)
            expected = self.simplify_parse_error(e)
            self.print_error(filename, line, col, "Expected %s" % expected)
            return False
        finally:
            self.input_sources = oldsrcs
            self.context.parsers.pop()
            self.context.filenames.pop()

    def parse_file(self, infile):
        currdir = os.getcwd()
        indir, infile = os.path.split(os.path.realpath(infile))
        try:
            os.chdir(indir)
            with open(infile) as f:
                src = f.read()
            return self.parse_string(
                src,
                grammar=program,
                filename=infile
            )
        finally:
            os.chdir(currdir)

    def include_file(self, infile, pos):
        currdir = os.getcwd()
        self.context.scope_push()
        if infile.startswith('!'):
            infile = infile[1:]
            src = resource_string(__name__, "incls/%s" % infile)
            src = src.decode("utf-8", errors="strict")
        elif self.sysincludes_only:
            line, col = self.context.parsers[-1].pos_to_linecol(pos)
            filename = self.context.filenames[-1]
            self.print_error(
                filename, line, col,
                "Non-system (!) includes disallowed."
            )
            return False
        else:
            indir, infile = os.path.split(os.path.realpath(infile))
            os.chdir(indir)
            with open(infile) as f:
                src = f.read()
        try:
            return self.parse_string(
                src,
                grammar=source_file,
                filename=infile
            )
        finally:
            self.context.scope_pop()
            os.chdir(currdir)


# vim: set ts=4 sw=4 et ai hlsearch nowrap :



#######################################################################
# Arpeggio PEG Grammar for MUV
#######################################################################

from __future__ import unicode_literals

from arpeggio import (
    Optional, ZeroOrMore, OneOrMore, EOF, Kwd
)
from arpeggio import RegExMatch as _


def comment_line():
    return _(r'//.*?$')


def comment_multi():
    return _(r'/\*(?ms).*?\*/')


def comment():
    return [comment_line, comment_multi]


def DIR_AUTHOR():
    return "$author"


def DIR_ECHO():
    return "$echo"


def DIR_ERROR():
    return "$error"


def DIR_INCLUDE():
    return "$include"


def DIR_LANGUAGE():
    return "$language"


def DIR_LIBVERSION():
    return "$libversion"


def DIR_NOTE():
    return "$note"


def DIR_PRAGMA():
    return "$pragma"


def DIR_VERSION():
    return "$version"


def DIR_WARN():
    return "$warn"


def KW_BREAK():
    return Kwd("break")


def KW_BY():
    return Kwd("by")


def KW_CASE():
    return Kwd("case")


def KW_CATCH():
    return Kwd("catch")


def KW_CONST():
    return Kwd("const")


def KW_CONTINUE():
    return Kwd("continue")


def KW_DEFAULT():
    return Kwd("default")


def KW_DEL():
    return Kwd("del")


def KW_DO():
    return Kwd("do")


def KW_ELSE():
    return Kwd("else")


def KW_EXTERN():
    return Kwd("extern")


def KW_FMTSTRING():
    return Kwd("fmtstring")


def KW_FOR():
    return Kwd("for")


def KW_FUNC():
    return Kwd("func")


def KW_IF():
    return Kwd("if")


def KW_IN():
    return Kwd("in")


def KW_INCLUDE():
    return Kwd("include")


def KW_MUF():
    return Kwd("muf")


def KW_MULTIPLE():
    return Kwd("multiple")


def KW_NAMESPACE():
    return Kwd("namespace")


def KW_PUBLIC():
    return Kwd("public")


def KW_PUSH():
    return Kwd("push")


def KW_RETURN():
    return Kwd("return")


def KW_SINGLE():
    return Kwd("single")


def KW_SWITCH():
    return Kwd("switch")


def KW_TOP():
    return Kwd("top")


def KW_TRY():
    return Kwd("try")


def KW_UNLESS():
    return Kwd("unless")


def KW_UNTIL():
    return Kwd("until")


def KW_USING():
    return Kwd("using")


def KW_VAR():
    return Kwd("var")


def KW_VOID():
    return Kwd("void")


def KW_WHILE():
    return Kwd("while")


def VARARG_MARKER():
    return "*"


def KEYVAL():
    return "=>"


def DOT():
    return "."


def COMMA():
    return ","


def COLON():
    return ":"


def SEMICOLON():
    return ";"


def PAREN():
    return "("


def ENDPAR():
    return ")"


def BRACE():
    return "{"


def ENDBRACE():
    return "}"


def BRACKET():
    return "["


def ENDBRACKET():
    return "]"


def ASGN():
    return _("=(?![=>])")


def ASGN_PLUS():
    return "+="


def ASGN_MINUS():
    return "-="


def ASGN_MULT():
    return "*="


def ASGN_DIV():
    return "/="


def ASGN_MOD():
    return "%="


def ASGN_POWER():
    return "**="


def ASGN_BITAND():
    return "&="


def ASGN_BITOR():
    return "|="


def ASGN_BITXOR():
    return "^="


def ASGN_BITLEFT():
    return "<<="


def ASGN_BITRIGHT():
    return ">>="


def OPER_PLUS():
    return _("\+(?![=+])")


def OPER_MINUS():
    return _("-(?![=-])")


def OPER_MULT():
    return _("\*(?![/*=])")


def OPER_DIV():
    return _("/(?![/*=])")


def OPER_MOD():
    return _("%(?!=)")


def OPER_POWER():
    return _("\*\*(?!=)")


def OPER_BITAND():
    return _("&(?![&=])")


def OPER_BITOR():
    return _("\|(?![|=])")


def OPER_BITXOR():
    return _(r'\^(?![=^])', str_repr='^')


def OPER_BITNOT():
    return "~"


def OPER_BITLEFT():
    return _("<<(?!=)")


def OPER_BITRIGHT():
    return _(">>(?!=)")


def OPER_INCR():
    return "++"


def OPER_DECR():
    return "--"


def OPER_TERNARY():
    return "?"


def LOGICAL_AND():
    return "&&"


def LOGICAL_OR():
    return "||"


def LOGICAL_XOR():
    return "^^"


def LOGICAL_NOT():
    return _("!(?!=)")


def COMP_EQ():
    return "=="


def COMP_NEQ():
    return "!="


def COMP_LT():
    return _(r"<(?![=<])")


def COMP_GT():
    return _(r">(?![=>])")


def COMP_LTE():
    return "<="


def COMP_GTE():
    return ">="


def COMP_IN():
    return "in"


def COMP_STREQ():
    return "eq"


def NS_SEP():
    return "::"


def identifier():
    return _(r'(?i)[a-z_]\w*')


def ns_parts():
    return Optional(NS_SEP), ZeroOrMore(identifier, NS_SEP)


def ATTR_IDENT():
    return identifier


def NS_IDENT():
    return ns_parts, identifier


def FUNC_IDENT():
    return ns_parts, identifier, Optional("?")


def DECL_VAR_IDENT():
    return identifier


def VAR_IDENT():
    return ns_parts, identifier


def CONST_IDENT():
    return ns_parts, identifier


def numsign():
    return ["+", "-"]


def bin_integer():
    return "0b", _(r'[01_]+')


def oct_integer():
    return "0o", _(r'[0-7_]+')


def dec_integer():
    return Optional("0d"), _(r'[0-9_]+(?![.eE])')


def hex_integer():
    return "0x", _(r'[0-9a-fA-F_]+')


def unsigned_integer():
    return [hex_integer, oct_integer, bin_integer, dec_integer]


def intnum():
    return Optional(numsign), unsigned_integer


def floatnum():
    return [
        _(r'[+-]?(\d+[.]\d*|\d*[.]\d+|\d+)[eE][+-]?[0-9]+'),
        _(r'[+-]?(\d+[.]\d*|\d*[.]\d+)'),
    ]


def number():
    return [floatnum, intnum]


def dbref():
    return "#", intnum


def raw_dquote_string():
    return [_('r".*?"')]


def raw_squote_string():
    return [_("r'.*?'")]


def raw_dquote3_string():
    return [_('(?ms)r""".*?"""')]


def raw_squote3_string():
    return [_("(?ms)r'''.*?'''")]


def dquote_string():
    return [_(r'"([^"\\]|\\.|\\$)*"')]


def squote_string():
    return [_(r"'([^'\\]|\\.|\\$)*'")]


def dquote3_string():
    return [_(r'(?ms)"""([^"\\]|\\.|"(?!""))*"""')]


def squote3_string():
    return [_(r"(?ms)'''([^'\\]|\\.|'(?!''))*'''")]


def raw_string_literal():
    return [
        raw_dquote3_string,
        raw_squote3_string,
        raw_dquote_string,
        raw_squote_string,
    ]


def esc_string_literal():
    return [
        dquote3_string,
        squote3_string,
        dquote_string,
        squote_string
    ]


def string_literal():
    return [raw_string_literal, esc_string_literal]


def declist():
    return DECL_VAR_IDENT, ZeroOrMore(COMMA, DECL_VAR_IDENT)


def argvar_list():
    return [
        (PAREN, ENDPAR),
        (PAREN, declist, Optional(VARARG_MARKER), ENDPAR)
    ]


def arglist():
    return assignment_expr, ZeroOrMore(COMMA, assignment_expr)


def paren_expr():
    return PAREN, expr, ENDPAR


def extern_type():
    return [KW_VOID, KW_SINGLE, KW_MULTIPLE]


def keyval_pair():
    return assignment_expr, KEYVAL, assignment_expr


def dictlist():
    return [
        KEYVAL,
        (keyval_pair, ZeroOrMore(COMMA, keyval_pair))
    ]


def compr_cond_if():
    return KW_IF, paren_expr


def compr_cond_unless():
    return KW_UNLESS, paren_expr


def compr_cond():
    return [compr_cond_if, compr_cond_unless]


def compr_for():
    return [for_expr, foreach_expr]


def raw_muf():
    return KW_MUF, PAREN, string_literal, ENDPAR


def stmt_raw_muf():
    return raw_muf, SEMICOLON


def gstmt_const():
    return KW_CONST, CONST_IDENT, ASGN, expr, SEMICOLON


def gstmt_using_ns():
    return KW_USING, KW_NAMESPACE, NS_IDENT, SEMICOLON


def gstmt_include():
    return KW_INCLUDE, string_literal, SEMICOLON


def gstmt_var_declinit():
    return KW_VAR, DECL_VAR_IDENT, ASGN, expr, SEMICOLON


def gstmt_var_declonly():
    return KW_VAR, DECL_VAR_IDENT, SEMICOLON


def gstmt_var():
    return [gstmt_var_declinit, gstmt_var_declonly]


def simple_extern():
    return KW_EXTERN, extern_type, FUNC_IDENT, argvar_list, SEMICOLON


def raw_muf_extern():
    return (
        KW_EXTERN, extern_type, FUNC_IDENT, argvar_list,
        ASGN, string_literal, SEMICOLON
    )


def opt_public():
    return Optional(KW_PUBLIC)


def gstmt_func():
    return opt_public, KW_FUNC, FUNC_IDENT, argvar_list, statements


def nsdecl():
    return KW_NAMESPACE, NS_IDENT, BRACE, global_statements, ENDBRACE


def global_statement():
    return [
        directive,
        stmt_raw_muf,
        gstmt_using_ns,
        gstmt_include,
        gstmt_const,
        gstmt_var,
        simple_extern,
        raw_muf_extern,
        gstmt_func
    ]


def variable():
    return VAR_IDENT


def declared_lvalue():
    return KW_VAR, variable


def indexed_lvalue():
    return variable, ZeroOrMore(index_part)


def lvalue():
    return [declared_lvalue, indexed_lvalue]


def tuple_parts():
    return COMP_LT, lvalue, ZeroOrMore(COMMA, lvalue), COMP_GT


def subscript():
    return BRACKET, expr, ENDBRACKET


def attribute():
    return DOT, ATTR_IDENT


def index_part():
    return [subscript, attribute]


def settable():
    return [tuple_parts, lvalue]


def stmt_return():
    return KW_RETURN, Optional(expr)


def stmt_break():
    return KW_BREAK


def stmt_continue():
    return KW_CONTINUE


def stmt_expr():
    return expr


def simple_statement():
    return [
        stmt_return,
        stmt_break,
        stmt_continue,
        stmt_expr,
    ]


def stmt_var():
    return KW_VAR, DECL_VAR_IDENT, SEMICOLON


def stmt_const():
    return KW_CONST, CONST_IDENT, ASGN, expr, SEMICOLON


def stmt_do_if():
    return simple_statement, KW_IF, paren_expr, SEMICOLON


def stmt_do_unless():
    return simple_statement, KW_UNLESS, paren_expr, SEMICOLON


def stmt_simple():
    return simple_statement, SEMICOLON


def stmt_if_else():
    return [
        (KW_IF, paren_expr, statement, KW_ELSE, statement),
        (KW_IF, paren_expr, statement),
    ]


def stmt_while():
    return KW_WHILE, paren_expr, statement


def stmt_until():
    return KW_UNTIL, paren_expr, statement


def stmt_do_while():
    return KW_DO, statement, KW_WHILE, paren_expr, SEMICOLON


def stmt_do_until():
    return KW_DO, statement, KW_UNTIL, paren_expr, SEMICOLON


def for_expr():
    return (
        KW_FOR, PAREN,
        lvalue, KW_IN,
        expr, KEYVAL, expr, Optional(KW_BY, expr),
        ENDPAR
    )


def foreach_expr():
    return (
        KW_FOR, PAREN,
        settable, Optional(KEYVAL, settable),
        KW_IN, expr,
        ENDPAR
    )


def stmt_for():
    return for_expr, statement


def stmt_foreach():
    return foreach_expr, statement


def stmt_try():
    return (
        KW_TRY, statement,
        KW_CATCH, PAREN, Optional(settable), ENDPAR,
        statement
    )


def function_name():
    return FUNC_IDENT


def usable_comparator():
    return [
        COMP_EQ,
        COMP_NEQ,
        COMP_LT,
        COMP_GT,
        COMP_LTE,
        COMP_GTE,
        COMP_IN,
        COMP_STREQ,
        "strcmp",
        "stringcmp",
    ]


def using_comparator():
    return KW_USING, usable_comparator


def using_raw_muf():
    return KW_USING, raw_muf


def using_function():
    return KW_USING, function_name


def expr_using():
    return [
        (expr, using_comparator),
        (expr, using_raw_muf),
        (expr, using_function),
        expr,
    ]


def case_clause():
    return KW_CASE, paren_expr, statement


def case_clauses():
    return OneOrMore(case_clause)


def default_clause():
    return KW_DEFAULT, statement


def switch_default():
    return Optional(default_clause)


def stmt_switch():
    return (
        KW_SWITCH, PAREN, expr_using, ENDPAR,
        BRACE, case_clauses, switch_default, ENDBRACE
    )


def statements():
    return BRACE, ZeroOrMore(statement), ENDBRACE


def statement():
    return [
        SEMICOLON,
        stmt_const,
        stmt_var,
        stmt_do_if,
        stmt_do_unless,
        stmt_simple,
        stmt_if_else,
        stmt_while,
        stmt_until,
        stmt_do_while,
        stmt_do_until,
        stmt_for,
        stmt_foreach,
        stmt_try,
        stmt_switch,
        statements,
    ]


def del_expr():
    return KW_DEL, PAREN, lvalue, ENDPAR


def push_expr():
    return KW_PUSH, PAREN, arglist, ENDPAR


def fmtstring_expr():
    return KW_FMTSTRING, PAREN, arglist, ENDPAR


def function_addr():
    return OPER_BITAND, FUNC_IDENT


def function_call():
    return FUNC_IDENT, PAREN, Optional(arglist), ENDPAR


def postfix_oper():
    return [OPER_INCR, OPER_DECR]


def unary_oper():
    return [
        OPER_INCR, OPER_DECR,
        OPER_PLUS, OPER_MINUS,
        OPER_BITNOT, LOGICAL_NOT
    ]


def power_oper():
    return OPER_POWER


def multiplicative_oper():
    return [OPER_MULT, OPER_DIV, OPER_MOD]


def additive_oper():
    return [OPER_PLUS, OPER_MINUS]


def shift_oper():
    return [OPER_BITLEFT, OPER_BITRIGHT]


def relational_oper():
    return [COMP_LTE, COMP_GTE, COMP_LT, COMP_GT, COMP_IN]


def equality_oper():
    return [COMP_EQ, COMP_NEQ, COMP_STREQ]


def assignment_oper():
    return [
        ASGN_MULT, ASGN_DIV, ASGN_MOD,
        ASGN_PLUS, ASGN_MINUS,
        ASGN_BITLEFT, ASGN_BITRIGHT,
        ASGN_BITAND, ASGN_BITXOR, ASGN_BITOR
    ]


def dict_comprehension():
    return BRACKET, compr_for, Optional(compr_cond), keyval_pair, ENDBRACKET


def list_comprehension():
    return BRACKET, compr_for, Optional(compr_cond), expr, ENDBRACKET


def dict_initialization():
    return BRACKET, dictlist, ENDBRACKET


def list_initialization():
    return [
        (BRACKET, ENDBRACKET),
        (BRACKET, arglist, ENDBRACKET)
    ]


def comprehension():
    return [
        dict_comprehension,
        list_comprehension,
        dict_initialization,
        list_initialization,
    ]


def leaf_variable():
    return variable


def leaf_expr():
    return [
        paren_expr,
        comprehension,
        number,
        dbref,
        string_literal,
        fmtstring_expr,
        KW_TOP,
        del_expr,
        push_expr,
        raw_muf,
        function_addr,
        function_call,
        leaf_variable,
    ]


def index_oper():
    return index_part


def addrcall_oper():
    return PAREN, Optional(arglist), ENDPAR


def subscript_expr():
    return leaf_expr, ZeroOrMore([index_oper, addrcall_oper])


def postfix_expr():
    return [
        (subscript_expr, postfix_oper),
        subscript_expr
    ]


def unary_expr():
    return [
        postfix_expr,
        (unary_oper, unary_expr)
    ]


def power_expr():
    return unary_expr, ZeroOrMore(power_oper, unary_expr)


def multiplicative_expr():
    return power_expr, ZeroOrMore(multiplicative_oper, power_expr)


def additive_expr():
    return multiplicative_expr, ZeroOrMore(additive_oper, multiplicative_expr)


def shift_expr():
    return additive_expr, ZeroOrMore(shift_oper, additive_expr)


def relational_expr():
    return shift_expr, ZeroOrMore(relational_oper, shift_expr)


def equality_expr():
    return relational_expr, ZeroOrMore(equality_oper, relational_expr)


def bitand_expr():
    return equality_expr, ZeroOrMore(OPER_BITAND, equality_expr)


def bitxor_expr():
    return bitand_expr, ZeroOrMore(OPER_BITXOR, bitand_expr)


def bitor_expr():
    return bitxor_expr, ZeroOrMore(OPER_BITOR, bitxor_expr)


def logical_and_expr():
    return bitor_expr, ZeroOrMore(LOGICAL_AND, bitor_expr)


def logical_xor_expr():
    return logical_and_expr, ZeroOrMore(LOGICAL_XOR, logical_and_expr)


def logical_or_expr():
    return logical_xor_expr, ZeroOrMore(LOGICAL_OR, logical_xor_expr)


def conditional_expr():
    return [
        (
            logical_or_expr,
            OPER_TERNARY, assignment_expr,
            COLON, assignment_expr
        ),
        logical_or_expr
    ]


def asgn_expr():
    return settable, ASGN, assignment_expr


def asgn_append_expr():
    return settable, BRACKET, ENDBRACKET, ASGN, assignment_expr


def asgn_oper_expr():
    return lvalue, assignment_oper, assignment_expr


def assignment_expr():
    return [
        asgn_expr,
        asgn_append_expr,
        asgn_oper_expr,
        conditional_expr
    ]


def expr():
    return assignment_expr, ZeroOrMore(COMMA, assignment_expr)


def dir_language():
    return DIR_LANGUAGE, string_literal


def dir_version():
    return DIR_VERSION, floatnum


def dir_libversion():
    return DIR_LIBVERSION, floatnum


def dir_author():
    return DIR_AUTHOR, string_literal


def dir_note():
    return DIR_NOTE, string_literal


def dir_echo():
    return DIR_ECHO, string_literal


def dir_pragma():
    return DIR_PRAGMA, string_literal


def dir_include():
    return DIR_INCLUDE, string_literal


def dir_error():
    return DIR_ERROR, string_literal


def dir_warn():
    return DIR_WARN, string_literal


def directive():
    return [
        dir_author,
        dir_echo,
        dir_error,
        dir_include,
        dir_language,
        dir_libversion,
        dir_note,
        dir_pragma,
        dir_version,
        dir_warn,
    ]


def global_statements():
    return ZeroOrMore([global_statement, nsdecl])


def source_file():
    return global_statements, EOF


def program():
    return [source_file]


# vim: set ts=4 sw=4 et ai hlsearch nowrap :



#######################################################################
# Utility classes
#######################################################################

from __future__ import unicode_literals
from __future__ import print_function

from pymuv.errors import MuvExceptionAlreadyDeclared


class FuncInfo(object):
    def __init__(
                self, pos, funcname,
                argcount=0,
                varargs=False,
                retcount=1,
                is_extern=False,
                code=None,
            ):
        self.position = pos
        self.funcname = funcname
        self.argcount = argcount
        self.varargs = varargs
        self.retcount = retcount
        self.is_extern = is_extern
        self.is_extern = is_extern
        self.code = code if code else funcname


class Scope(object):
    def __init__(self):
        self.variables = {}
        self.constants = {}

    def lookup_constant(self, name):
        if name not in self.constants:
            return None
        return self.constants[name]

    def lookup_variable(self, name):
        if name not in self.variables:
            return None
        return self.variables[name]

    def declare_constant(self, name, val):
        if name in self.constants:
            raise MuvExceptionAlreadyDeclared()
        self.constants[name] = val

    def declare_variable(self, name, realname):
        if name in self.variables:
            raise MuvExceptionAlreadyDeclared()
        self.variables[name] = realname


class Context(object):
    def __init__(self):
        self.parsers = []
        self.filenames = []
        self.debug = False
        self.scopes = [Scope()]
        self.func_realvars = {}
        self.curr_namespace = '::'
        self.using_namespaces = []
        self.global_consts = {}
        self.global_vars = {}
        self.init_statements = []
        self.functions = {}
        self.last_function = None
        self.last_func_pos = None
        self.assign_level = 0
        self.optimization_level = 1

        # Standard Global Variables
        self.declare_global_var("me", sys=True)
        self.declare_global_var("loc", sys=True)
        self.declare_global_var("trigger", sys=True)
        self.declare_global_var("command", sys=True)

        # Global Initialiations
        self.add_init('"me" match me ! me @ location loc ! trig trigger !')

        # Standard Builtin Functions
        self.declare_function(
            "abort", 1, retcount=0, is_extern=True, code="abort")
        self.declare_function(
            "throw", 1, retcount=0, is_extern=True, code="abort")
        self.declare_function(
            "tell", 1, retcount=0, is_extern=True, code="me @ swap notify")
        self.declare_function(
            "count", 1, retcount=1, is_extern=True, code="array_count")
        self.declare_function(
            "cat", 1, varargs=True, retcount=1, is_extern=True,
            code="array_interpret"
        )
        self.declare_function(
            "haskey", 2, retcount=1, is_extern=True,
            code="swap 1 array_make array_extract"
        )

        # Global Constants
        self.declare_global_const("true", "1")
        self.declare_global_const("false", "0")

    # Misc ###################################################

    def scope_push(self):
        self.scopes.append(Scope())

    def scope_pop(self):
        self.scopes.pop()

    def add_init(self, expstr):
        self.init_statements.append(expstr)

    # NameSpaces #############################################

    def set_namespace(self, ns):
        if ns.startswith('::'):
            self.curr_namespace = ns
        else:
            if self.curr_namespace != "::":
                self.curr_namespace += '::'
            self.curr_namespace += ns

    def use_namespace(self, ns):
        if not ns.startswith('::'):
            oldns = self.curr_namespace
            if oldns == "::":
                oldns = ''
            ns = '%s::%s' % (oldns, ns)
        self.using_namespaces.append(ns)

    def canonical_name(self, name):
        if name.startswith('::'):
            return name
        ns = self.curr_namespace
        if ns == "::":
            ns = ""
        out = "{ns}::{name}".format(ns=ns, name=name)
        while '::::' in ns:
            ns = ns.replace('::::', '::')
        return out

    def actualize_name(self, name):
        name = self.canonical_name(name)
        name = name.strip(':').replace('::', '__')
        if "__" not in name:
            name = "_" + name
        return name

    def possible_canonicals(self, name):
        out = []
        if name.startswith('::'):
            return [name]
        for ns in self.using_namespaces:
            if ns == '::':
                ns = ''
            canonical = "%s::%s" % (ns, name)
            out.append(canonical)
        ns = self.curr_namespace
        while True:
            if ns == '::':
                ns = ''
            canonical = "%s::%s" % (ns, name)
            out.append(canonical)
            if not ns:
                break
            ns = ns.rsplit('::', 1)[0]
        return out

    # Constants ##############################################

    def lookup_constant(self, name):
        for scope in reversed(self.scopes):
            found = scope.lookup_constant(name)
            if found:
                return found
        for name in self.possible_canonicals(name):
            if name in self.global_consts:
                return self.global_consts[name]
        return None

    def declare_constant(self, name, val):
        self.scopes[-1].declare_constant(name, val)

    def declare_global_const(self, name, val):
        name = self.canonical_name(name)
        self.global_consts[name] = val

    # Variables ##############################################

    def lookup_variable(self, name):
        for scope in reversed(self.scopes):
            found = scope.lookup_variable(name)
            if found:
                return found
        for name in self.possible_canonicals(name):
            if name in self.global_vars:
                return self.global_vars[name]
        return None

    def declare_variable(self, name):
        realname = name
        if "__" not in realname:
            realname = "_%s" % realname
        for i in range(2, 99):
            if realname not in self.func_realvars:
                break
            realname = "%s%d" % (name, i)
            if "__" not in realname:
                realname = "_%s" % realname
        self.scopes[-1].declare_variable(name, realname)
        self.func_realvars[realname] = name
        return realname

    def declare_global_var(self, name, sys=False):
        realname = name
        name = self.canonical_name(name)
        if not sys:
            realname = self.actualize_name(name)
        self.global_vars[name] = realname
        return realname

    # Functions ##############################################

    def lookup_function(self, name):
        for name in self.possible_canonicals(name):
            if name in self.functions:
                return self.functions[name]
        return None

    def declare_function(
                self, name, argcount,
                varargs=False,
                retcount=1,
                is_extern=False,
                code=None,
                is_public=False,
                pos=None
            ):
        canonical = self.canonical_name(name)
        actual = self.actualize_name(name)
        if is_public and actual.startswith("_"):
            actual = actual[1:]
        if not code:
            code = actual
        self.functions[canonical] = FuncInfo(
            pos, actual,
            argcount=argcount,
            varargs=varargs,
            retcount=retcount,
            is_extern=is_extern,
            code=code
        )
        self.last_function = actual
        self.last_func_pos = pos
        return actual

    def reset_function_local_data(self):
        self.scopes = [Scope()]
        self.func_realvars = {}


# vim: set ts=4 sw=4 et ai hlsearch nowrap :



class MuvError(Exception):
    def __init__(self, msg, position=0):
        self.position = position
        super(MuvError, self).__init__(msg)


class MuvExceptionAlreadyDeclared(Exception):
    pass


# vim: set ts=4 sw=4 et ai hlsearch nowrap :



from pymuv.errors import MuvError


class LValue(object):
    def __init__(self, varname, indexing, pos, declare=False, readonly=False):
        self.varname = varname
        self.indexing = indexing
        self.declare = declare
        self.position = pos
        self.readonly = readonly

    def get_expr(self, ctx):
        """
        Returns the MUF needed to get the contents of the lvalue.
        Returned MUF will push the contained value onto the stack.
        """
        varname = ctx.lookup_variable(self.varname)
        if varname is None:
            val = ctx.lookup_constant(self.varname)
            if val:
                try:
                    return val.generate_code(ctx)
                except AttributeError:
                    return val
            raise MuvError(
                "Undeclared identifier '%s'." % self.varname,
                position=self.position
            )
        if len(self.indexing) == 0:
            return "{var} @".format(
                var=varname,
            )
        if len(self.indexing) == 1:
            return "{var} @ {idx} []".format(
                var=varname,
                idx=self.indexing[0],
            )
        return (
            "{var} @ {{ {idx} }}list array_nested_get".format(
                var=varname,
                idx=" ".join(str(x) for x in self.indexing),
            )
        )

    def set_expr(self, ctx):
        """
        Returns the MUF needed to do an assign on the lvalue. (=)
        Returned MUF expects a value to be on the stack.
        """
        if self.readonly:
            raise MuvError(
                "Cannot assign value to constant '%s'." % self.varname,
                position=self.position
            )
        if self.declare:
            varname = ctx.declare_variable(self.varname)
            if len(self.indexing) == 0:
                return "var! {var}".format(var=varname)
        else:
            varname = ctx.lookup_variable(self.varname)
            if varname is None:
                raise MuvError(
                    "Undeclared identifier '%s'." % self.varname,
                    position=self.position
                )
        if len(self.indexing) == 0:
            if ctx.assign_level > 1:
                return "dup {var} !".format(var=varname)
            else:
                return "{var} !".format(var=varname)
        if len(self.indexing) == 1:
            if ctx.assign_level > 1:
                fmt = "dup {var} @ {idx} ->[] {var} !"
            else:
                fmt = "{var} @ {idx} ->[] {var} !"
            return fmt.format(
                var=varname,
                idx=self.indexing[0].generate_code(ctx),
            )
        if ctx.assign_level > 1:
            fmt = "dup {var} @ {{ {idx} }}list array_nested_set {var} !"
        else:
            fmt = "{var} @ {{ {idx} }}list array_nested_set {var} !"
        return fmt.format(
            var=varname,
            idx=" ".join(x.generate_code(ctx) for x in self.indexing)
        )

    def oper_set_expr(self, ctx, oper, val):
        """
        Returns the MUF needed to do an oper-assign on the lvalue. (+=, etc.)
        """
        if self.readonly:
            raise MuvError(
                "Cannot assign value to constant '%s'." % self.varname,
                position=self.position
            )
        varname = ctx.lookup_variable(self.varname)
        if varname is None:
            raise MuvError(
                "Undeclared identifier '%s'." % self.varname,
                position=self.position
            )
        if len(self.indexing) == 0:
            if ctx.assign_level > 1:
                fmt = "{var} @ {val} {oper} dup {var} !"
            else:
                fmt = "{var} @ {val} {oper} {var} !"
            return fmt.format(
                var=varname,
                oper=oper,
                val=val.generate_code(ctx),
            )
        if len(self.indexing) == 1:
            if ctx.assign_level > 1:
                fmt = (
                    "{var} @ {idx} "
                    "over over [] {val} {oper} "
                    "dup -4 rotate "
                    "rot rot ->[] {var} !"
                )
            else:
                fmt = (
                    "{var} @ {idx} "
                    "over over [] {val} {oper} "
                    "rot rot ->[] {var} !"
                )
            return fmt.format(
                var=varname,
                oper=oper,
                val=val.generate_code(ctx),
                idx=self.indexing[0].generate_code(ctx),
            )
        if ctx.assign_level > 1:
            fmt = (
                "{var} @ {{ {idx} }}list "
                "over over array_nested_get {val} {oper} "
                "dup -4 rotate "
                "rot rot array_nested_set {var} !"
            )
        else:
            fmt = (
                "{var} @ {{ {idx} }}list "
                "over over array_nested_get {val} {oper} "
                "rot rot array_nested_set {var} !"
            )
        return fmt.format(
            var=varname,
            oper=oper,
            val=val.generate_code(ctx),
            idx=" ".join(
                x.generate_code(ctx)
                for x in self.indexing
            )
        )

    def unary_set_expr(self, ctx, oper, postoper=False):
        """
        Returns the MUF needed to do an unary operation on the lvalue. (++, --.)
        """
        if self.readonly:
            raise MuvError(
                "Cannot increment or decrement constant '%s'." % self.varname,
                position=self.position
            )
        varname = ctx.lookup_variable(self.varname)
        if varname is None:
            raise MuvError(
                "Undeclared identifier '%s'." % self.varname,
                position=self.position
            )
        if len(self.indexing) == 0:
            if postoper:
                fmt = "{var} @ {var} {oper}"
            else:
                fmt = "{var} dup {oper} @"
            return fmt.format(var=varname, oper=oper)
        if len(self.indexing) == 1:
            if postoper:
                fmt = (
                    "{idx} {var} @ "
                    "dup 3 pick [] "
                    "dup -4 rotate {oper} "
                    "swap rot ->[] {var} !"
                )
            else:
                fmt = (
                    "{idx} {var} @ "
                    "dup 3 pick [] {oper} "
                    "dup -4 rotate "
                    "swap rot ->[] {var} !"
                )
            return fmt.format(
                var=varname,
                oper=oper,
                idx=self.indexing[0].generate_code(ctx),
            )
        if postoper:
            fmt = (
                "{{ {idx} }}list {var} @ "
                "dup 3 pick array_nested_get "
                "dup -4 rotate {oper} "
                "swap rot array_nested_set {var} !"
            )
        else:
            fmt = (
                "{{ {idx} }}list {var} @ "
                "dup 3 pick array_nested_get {oper} "
                "dup -4 rotate "
                "swap rot array_nested_set {var} !"
            )
        return fmt.format(
            var=varname,
            oper=oper,
            idx=" ".join(
                x.generate_code(ctx)
                for x in self.indexing
            )
        )

    def del_expr(self, ctx):
        """
        Returns the MUF needed to delete a given lvalue. (ie: array item)
        Returned MUF will set a bare variable to 0, and will remove the
        given indexed item for an indexed array or dictionary.
        """
        if self.readonly:
            raise MuvError(
                "Cannot assign value to constant '%s'." % self.varname,
                position=self.position
            )
        varname = ctx.lookup_variable(self.varname)
        if varname is None:
            raise MuvError(
                "Undeclared identifier '%s'." % self.varname,
                position=self.position
            )
        if len(self.indexing) == 0:
            return "0 {var} !".format(var=varname)
        if len(self.indexing) == 1:
            return "{var} @ {idx} array_delitem {var} !".format(
                var=varname,
                idx=self.indexing[0].generate_code(ctx),
            )
        return "{var} @ {{ {idx} }}list array_nested_del {var} !".format(
            var=varname,
            idx=" ".join(x.generate_code(ctx) for x in self.indexing),
        )


# vim: set ts=4 sw=4 et ai hlsearch nowrap :



#!/usr/bin/env python3

from __future__ import unicode_literals
from __future__ import print_function

import sys
import argparse

from pymuv.parsefile import MuvParser


def main():
    parser = argparse.ArgumentParser(prog='pymuv')
    parser.add_argument("-d", "--debug",
                        help="Add debugging code/markers to output MUF.",
                        action="store_true")
    parser.add_argument("-c", "--check",
                        help="Don't output code.  Just check for errors.",
                        action="store_true")
    parser.add_argument("-n", "--no-optimization",
                        help="Disable code optimization.",
                        action="store_true")
    parser.add_argument("-w", "--wrapper", type=str, default="",
                        help="Wrap code for upload into given program.")
    parser.add_argument("-s", "--sysincludes-only",
                        help="Only allow system ! includes.",
                        action="store_true")
    parser.add_argument("-o", "--output", type=str, default="",
                        help="File to write MUF output to.")
    parser.add_argument('infile', help='Input MUV sourcecode file.')
    opts = parser.parse_args()

    muvparser = MuvParser()
    muvparser.set_debug(opts.debug)
    muvparser.optimization_level = 0 if opts.no_optimization else 1
    muvparser.sysincludes_only = opts.sysincludes_only
    muvparser.wrapper_program = opts.wrapper
    muvparser.parse_file(opts.infile)

    if not opts.check:
        try:
            outstream = sys.stdout
            if opts.output:
                outstream = open(opts.output, "w")
            if not muvparser.error_found:
                print(muvparser.output, file=outstream)
        finally:
            if opts.output:
                outstream.close()
    sys.exit(-1 if muvparser.error_found else 0)


if __name__ == "__main__":
    main()


# vim: set ts=4 sw=4 et ai hlsearch nowrap :



from __future__ import unicode_literals

from pymuv.errors import MuvError


class Settable(object):
    def __init__(self, *lvalues):
        self.settables = lvalues[1:-1]

    def get_expr(self, ctx):
        raise MuvError(
            "Cannot use tuple in right side expression.",
            position=self.settables[0].position
        )

    def set_expr(self, ctx):
        cnt = len(self.settables)
        if ctx.assign_level > 1:
            fmt = "dup 0 {cnt} [..] array_vals pop\n{exprs}"
        else:
            fmt = "0 {cnt} [..] array_vals pop\n{exprs}"
        out = fmt.format(
            cnt=cnt-1,
            exprs=" ".join(
                x.set_expr(ctx)
                for x in reversed(list(self.settables))
            )
        )
        return out

    def oper_set_expr(self, ctx, oper, val):
        raise MuvError(
            "Cannot use tuple in left side of operator-assignment expression.",
            position=self.settables[0].position
        )

    def unary_set_expr(self, ctx, oper, postoper=False):
        raise MuvError(
            "Cannot use tuple with increment/decrement operators.",
            position=self.settables[0].position
        )

    def del_expr(self, ctx):
        raise MuvError(
            "Cannot use tuple in del expression.",
            position=self.settables[0].position
        )


# vim: set ts=4 sw=4 et ai hlsearch nowrap :



#!/usr/bin/env python

import os
import glob
from setuptools import setup

VERSION = "0.9.2"


def find_data_files(source, target, patterns):
    """
    Locates the specified data-files and returns the matches
    in a data_files compatible format.

    source is the root of the source data tree.
        Use '' or '.' for current directory.
    target is the root of the target data tree.
        Use '' or '.' for the distribution directory.
    patterns is a sequence of glob-patterns for the
        files you want to copy.
    """
    if glob.has_magic(source) or glob.has_magic(target):
        raise ValueError("Magic not allowed in src, target")
    ret = {}
    for pattern in patterns:
        pattern = os.path.join(source, pattern)
        for filename in glob.glob(pattern):
            if os.path.isfile(filename):
                targetpath = os.path.join(
                    target, os.path.relpath(filename, source)
                )
                path = os.path.dirname(targetpath)
                ret.setdefault(path, []).append(filename)
    return sorted(ret.items())


with open('README.rst') as f:
    LONG_DESCR = f.read()


setup(
    name='pymuv',
    version=VERSION,
    description='A python implementation of the MUV 2.0 to fb6 MUF compiler.',
    long_description=LONG_DESCR,
    author='Revar Desmera',
    author_email='revarbat@gmail.com',
    url='https://github.com/revarbat/pymuv',
    download_url='https://github.com/revarbat/pymuv/archive/master.zip',
    packages=['pymuv'],
    package_data={'pymuv': ["incls/fb6/*"]},
    license='MIT License',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Compilers',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='muv muf muck compiler',
    entry_points={
        'console_scripts': ['pymuv=pymuv:main'],
    },
    install_requires=[
        'setuptools',
        'Arpeggio>=1.5',
        'six>=1.10.0',
    ],
)



#######################################################################
# Syntactic classes
#######################################################################

import re
import textwrap

from pymuv.errors import MuvError


def str_or_expr(x, ctx):
    if isinstance(x, MuvNode):
        return x.generate_code(ctx)
    return x


class MuvNode(object):
    def __init__(self, typ, pos):
        self.typename = typ
        self.position = pos
        self.children = []

    def append(self, child):
        self.children.append(child)

    def generate_code(self, ctx):
        out = " ".join(
            child.generate_code(ctx) for child in self.children
        )
        return out

    def indent(self, txt, ind='    '):
        return "\n".join((ind + l if l else l) for l in txt.split('\n'))

    def annotate(self, ctx, txt, force=False, pos=None, lineoff=0):
        if ctx.debug:
            if not pos:
                pos = self.position
            line, col = ctx.parsers[-1].pos_to_linecol(pos)
            if force or len(ctx.filenames) == 1:
                txt = "(MUV:L%d) %s" % (line + lineoff, txt)
        return txt

    def strip_extra_annotations(self, txt):
        out = ''
        txt = re.sub(r'(\(MUV:L\d+\) )+(\(MUV:L\d+\) )', r'\2', txt)
        prevline = -2
        while txt:
            m = re.match(r'(?ms)(.*?)\(MUV:L(\d+)\) (.*)', txt)
            if not m:
                out += txt
                txt = ''
                break
            out += m.group(1)
            line = m.group(2)
            txt = m.group(3)
            if line != prevline:
                prevline = line
                out += "(MUV:L%s) " % line
        return out


class MuvNodeNull(MuvNode):
    def __init__(self, pos):
        super(MuvNodeNull, self).__init__("NULL", pos)

    def generate_code(self, ctx):
        return ''


class MuvNodeInteger(MuvNode):
    def __init__(self, pos, val):
        super(MuvNodeInteger, self).__init__("INTEGER", pos)
        self.value = val
        self.valtype = "integer"

    def generate_code(self, ctx):
        return self.annotate(ctx, str(self.value))


class MuvNodeFloat(MuvNode):
    def __init__(self, pos, val):
        super(MuvNodeFloat, self).__init__("FLOAT", pos)
        self.value = val
        self.valtype = "float"

    def generate_code(self, ctx):
        return self.annotate(ctx, str(self.value))


class MuvNodeString(MuvNode):
    def __init__(self, pos, val):
        super(MuvNodeString, self).__init__("STRING", pos)
        self.value = val
        self.valtype = "string"

    def generate_code(self, ctx):
        out = self.value.replace('\x5c', '\x5c\x5c')
        out = out.replace('"', '\x5c"')
        out = out.replace('\r', '\x5cr')
        out = out.replace('\n', '\x5cr')
        out = out.replace('\x1b', '\x5c[')
        out = '"%s"' % out
        return self.annotate(ctx, out)


class MuvNodeDbRef(MuvNode):
    def __init__(self, pos, val):
        super(MuvNodeDbRef, self).__init__("DBREF", pos)
        self.value = val
        self.valtype = "dbref"

    def generate_code(self, ctx):
        return "#%d" % self.value


class MuvNodeKeyValPair(MuvNode):
    def __init__(self, pos, key, val):
        super(MuvNodeKeyValPair, self).__init__("KEYVAL_PAIR", pos)
        self.key = key
        self.val = val

    def generate_code(self, ctx):
        out = "{key} {val}".format(
            key=self.key.generate_code(ctx),
            val=self.val.generate_code(ctx),
        )
        return self.annotate(ctx, out)


class MuvNodeGlobalVar(MuvNode):
    def __init__(self, pos, name, initval=None):
        super(MuvNodeGlobalVar, self).__init__("GLOBAL_VAR", pos)
        self.name = name
        self.initval = initval

    def generate_code(self, ctx):
        realname = ctx.declare_global_var(self.name)
        if self.initval is not None:
            init = "{expr} {var} !".format(
                expr=self.initval.generate_code(ctx),
                var=realname
            )
            ctx.add_init(init)
        out = 'lvar %s' % realname
        return self.annotate(ctx, out)


class MuvNodeGlobalConst(MuvNode):
    def __init__(self, pos, name, value):
        super(MuvNodeGlobalConst, self).__init__("GLOBAL_CONST", pos)
        self.name = name
        self.value = value

    def generate_code(self, ctx):
        ctx.declare_global_const(self.name, self.value)
        return ''


class MuvNodeFuncConst(MuvNode):
    def __init__(self, pos, name, value):
        super(MuvNodeFuncConst, self).__init__("FUNC_CONST", pos)
        self.name = name
        self.value = value

    def generate_code(self, ctx):
        ctx.declare_constant(self.name, self.value)
        return ''


class MuvNodeFuncVar(MuvNode):
    def __init__(self, pos, name):
        super(MuvNodeFuncVar, self).__init__("FUNC_VAR", pos)
        self.name = name

    def generate_code(self, ctx):
        realname = ctx.declare_variable(self.name)
        out = "var {var}".format(var=realname)
        return self.annotate(ctx, out)


class MuvNodeVarFetch(MuvNode):
    def __init__(self, pos, var):
        super(MuvNodeVarFetch, self).__init__("VAR_FETCH", pos)
        self.var = var

    def generate_code(self, ctx):
        out = self.var.get_expr(ctx)
        return self.annotate(ctx, out)


class MuvNodeVarAssign(MuvNode):
    def __init__(self, pos, var, expr):
        super(MuvNodeVarAssign, self).__init__("VAR_ASSIGN", pos)
        self.var = var
        self.expr = expr

    def generate_code(self, ctx):
        ctx.assign_level += 1
        out = "{expr} {setexp}".format(
            expr=self.expr.generate_code(ctx),
            setexp=self.var.set_expr(ctx),
        )
        out = self.annotate(ctx, out)
        ctx.assign_level -= 1
        return out


class MuvNodeVarOperAssign(MuvNode):
    def __init__(self, pos, var, oper, expr):
        super(MuvNodeVarOperAssign, self).__init__("VAR_OPER_ASSIGN", pos)
        self.var = var
        self.oper = oper
        self.expr = expr

    def generate_code(self, ctx):
        ctx.assign_level += 1
        out = self.var.oper_set_expr(ctx, self.oper, self.expr)
        out = self.annotate(ctx, out)
        ctx.assign_level -= 1
        return out


class MuvNodeVarDel(MuvNode):
    def __init__(self, pos, var):
        super(MuvNodeVarDel, self).__init__("VAR_DEL", pos)
        self.var = var

    def generate_code(self, ctx):
        out = self.var.del_expr(ctx)
        return self.annotate(ctx, out)


class MuvNodePostfixExpr(MuvNode):
    def __init__(self, pos, expr, oper):
        super(MuvNodePostfixExpr, self).__init__("POSTFIX_EXPR", pos)
        self.expr = expr
        self.oper = oper

    def generate_code(self, ctx):
        if isinstance(self.expr, MuvNodeVarFetch):
            lval = self.expr.var
            ctx.assign_level += 1
            out = lval.unary_set_expr(ctx, self.oper, postoper=True)
            ctx.assign_level -= 1
        else:
            out = "{expr} {oper}".format(
                expr=self.expr.generate_code(ctx),
                oper=self.oper,
            )
        return self.annotate(ctx, out)


class MuvNodePrefixExpr(MuvNode):
    def __init__(self, pos, oper, expr):
        super(MuvNodePrefixExpr, self).__init__("PREFIX_EXPR", pos)
        self.oper = oper
        self.expr = expr

    def generate_code(self, ctx):
        if self.oper in ['+', '-']:
            out = "0 {expr} {oper}".format(
                expr=self.expr.generate_code(ctx),
                oper=self.oper,
            )
        elif isinstance(self.expr, MuvNodeVarFetch):
            lval = self.expr.var
            ctx.assign_level += 1
            out = lval.unary_set_expr(ctx, self.oper)
            ctx.assign_level -= 1
        else:
            out = "{expr} {oper}".format(
                expr=self.expr.generate_code(ctx),
                oper=self.oper,
            )
        return self.annotate(ctx, out)


class MuvNodeBinaryExpr(MuvNode):
    def __init__(self, pos, expr1, oper, expr2):
        super(MuvNodeBinaryExpr, self).__init__("BINARY_EXPR", pos)
        self.expr1 = expr1
        self.oper = oper
        self.expr2 = expr2

    def generate_code(self, ctx):
        ctx.assign_level += 1
        out = "{expr1} {expr2} {oper}".format(
            expr1=self.expr1.generate_code(ctx),
            oper=self.oper,
            expr2=self.expr2.generate_code(ctx),
        )
        ctx.assign_level -= 1
        return self.annotate(ctx, out)


class MuvNodeLogicalOr(MuvNode):
    def __init__(self, pos, expr1, expr2):
        super(MuvNodeLogicalOr, self).__init__("OR", pos)
        self.expr1 = expr1
        self.expr2 = expr2

    def generate_code(self, ctx):
        out = (
            "{expr1} dup not if pop\n"
            "{expr2}\n"
            "then"
        ).format(
            expr1=self.expr1.generate_code(ctx),
            expr2=self.indent(self.expr2.generate_code(ctx)),
        )
        return self.annotate(ctx, out)


class MuvNodeLogicalAnd(MuvNode):
    def __init__(self, pos, expr1, expr2):
        super(MuvNodeLogicalAnd, self).__init__("AND", pos)
        self.expr1 = expr1
        self.expr2 = expr2

    def generate_code(self, ctx):
        out = (
            "{expr1} dup if pop\n"
            "{expr2}\n"
            "then"
        ).format(
            expr1=self.expr1.generate_code(ctx),
            expr2=self.indent(self.expr2.generate_code(ctx)),
        )
        return self.annotate(ctx, out)


class MuvNodeExprList(MuvNode):
    def __init__(self, pos, *children):
        super(MuvNodeExprList, self).__init__("STATEMENTS", pos)
        self.children = list(children)

    def generate_code(self, ctx):
        out = " ".join(
            str_or_expr(child, ctx) for child in self.children
        )
        return self.annotate(ctx, out)


class MuvNodeStatements(MuvNode):
    def __init__(self, pos, *children):
        super(MuvNodeStatements, self).__init__("STATEMENTS", pos)
        self.children = []
        for child in children:
            if isinstance(child, MuvNodeStatements):
                for subchild in child.children:
                    self.children.append(subchild)
            else:
                self.children.append(child)

    def generate_code(self, ctx):
        ctx.scope_push()
        statements = [
            str_or_expr(child, ctx)
            for child in self.children
        ]
        out = "\n".join(x for x in statements if x)
        ctx.scope_pop()
        return self.annotate(ctx, out)


class MuvNodeCommand(MuvNode):
    def __init__(self, pos, cmd, *children):
        super(MuvNodeCommand, self).__init__("COMMAND", pos)
        self.command = cmd
        self.children = children

    def generate_code(self, ctx):
        ctx.assign_level += 1
        outlist = [child.generate_code(ctx) for child in self.children]
        outlist.append(self.command)
        out = " ".join(x for x in outlist if x)
        ctx.assign_level -= 1
        return self.annotate(ctx, out)


class MuvNodeWrappedExpr(MuvNode):
    def __init__(self, pos, pfx, sfx, *children):
        super(MuvNodeWrappedExpr, self).__init__("WRAPPED", pos)
        self.prefix = pfx
        self.suffix = sfx
        self.children = children

    def generate_code(self, ctx):
        bodyparts = [
            item for item in (
                x.generate_code(ctx) for x in self.children
            ) if item
        ]
        body = ""
        currlen = 0
        for item in bodyparts:
            itemlen = len(item)
            if currlen:
                body += " "
                currlen += 1
            body += item
            if currlen + itemlen > 40:
                body += "\n"
                currlen = 0
            else:
                currlen += itemlen
        if body.endswith('\n'):
            body = body[:-1]
        if not body:
            out = "{0} {1}".format(self.prefix, self.suffix)
        elif "\n" in body:
            body = self.indent(body)
            out = "{0}\n{1}\n{2}".format(self.prefix, body, self.suffix)
        else:
            out = "{0} {1} {2}".format(self.prefix, body, self.suffix)
        return self.annotate(ctx, out)


class MuvNodeIfElse(MuvNode):
    def __init__(
            self, pos,
            cond=None,
            ifclause=None,
            elseclause=None
            ):
        super(MuvNodeIfElse, self).__init__("IF", pos)
        self.conditional = cond
        self.ifclause = ifclause
        self.elseclause = elseclause

    def generate_code(self, ctx):
        ctx.scope_push()
        if self.elseclause:
            out = (
                "{cond} if\n"
                "{body}\n"
                "else\n"
                "{els}\n"
                "then"
            ).format(
                cond=self.conditional.generate_code(ctx),
                body=self.indent(self.ifclause.generate_code(ctx)),
                els=self.indent(self.elseclause.generate_code(ctx)),
            )
        else:
            out = (
                "{cond} if\n"
                "{body}\n"
                "then"
            ).format(
                cond=self.conditional.generate_code(ctx),
                body=self.indent(self.ifclause.generate_code(ctx)),
            )
        ctx.scope_pop()
        return self.annotate(ctx, out)


class MuvNodeForLoop(MuvNode):
    def __init__(
            self, pos,
            var=None,
            start=None,
            end=None,
            stride=None,
            body=None
            ):
        super(MuvNodeForLoop, self).__init__("FOR", pos)
        self.var = var
        self.start = start
        self.end = end
        self.stride = stride

    def generate_code(self, ctx):
        ctx.scope_push()
        out = (
            "{start} {end} {stride} for\n"
            "{vexp}\n"
            "{body}\n"
            "repeat"
        ).format(
            start=self.start.generate_code(ctx),
            end=self.end.generate_code(ctx),
            stride=self.stride.generate_code(ctx),
            vexp=self.indent(self.var.set_expr(ctx)),
            body=self.indent(self.body.generate_code(ctx)),
        )
        ctx.scope_pop()
        return self.annotate(ctx, out)


class MuvNodeForEachLoop(MuvNode):
    def __init__(self, pos, expr=None, keyvar=None, valvar=None, body=None):
        super(MuvNodeForEachLoop, self).__init__("FOREACH", pos)
        self.expr = expr
        self.keyvar = keyvar
        self.valvar = valvar
        self.body = body

    def generate_code(self, ctx):
        ctx.scope_push()
        out = (
            "{expr} foreach\n"
            "{vset} {kset}\n"
            "{body}\n"
            "repeat"
        ).format(
            expr=self.expr.generate_code(ctx),
            vset=self.indent(self.valvar.set_expr(ctx)),
            kset=self.keyvar.set_expr(ctx) if self.keyvar else "pop",
            body=self.indent(self.body.generate_code(ctx)),
        )
        ctx.scope_pop()
        return self.annotate(ctx, out)


class MuvNodeWhile(MuvNode):
    def __init__(self, pos, cond=None, body=None):
        super(MuvNodeWhile, self).__init__("WHILE", pos)
        self.conditional = cond
        self.body = body

    def generate_code(self, ctx):
        ctx.scope_push()
        out = (
            "begin\n"
            "{cond}\n"
            "while\n"
            "{body}\n"
            "repeat"
        ).format(
            cond=self.indent(self.conditional.generate_code(ctx)),
            body=self.indent(self.body.generate_code(ctx)),
        )
        ctx.scope_pop()
        return self.annotate(ctx, out)


class MuvNodeUntil(MuvNode):
    def __init__(self, pos, cond=None, body=None):
        super(MuvNodeUntil, self).__init__("UNTIL", pos)
        self.conditional = cond
        self.body = body

    def generate_code(self, ctx):
        ctx.scope_push()
        out = (
            "begin\n"
            "{cond} not\n"
            "while\n"
            "{body}\n"
            "repeat"
        ).format(
            cond=self.indent(self.conditional.generate_code(ctx)),
            body=self.indent(self.body.generate_code(ctx)),
        )
        ctx.scope_pop()
        return self.annotate(ctx, out)


class MuvNodeDoWhile(MuvNode):
    def __init__(self, pos, body=None, cond=None):
        super(MuvNodeDoWhile, self).__init__("DO_WHILE", pos)
        self.body = body
        self.conditional = cond

    def generate_code(self, ctx):
        ctx.scope_push()
        out = (
            "begin\n"
            "{body}\n"
            "{cond} not\n"
            "until"
        ).format(
            cond=self.indent(self.conditional.generate_code(ctx)),
            body=self.indent(self.body.generate_code(ctx)),
        )
        ctx.scope_pop()
        return self.annotate(ctx, out)


class MuvNodeDoUntil(MuvNode):
    def __init__(self, pos, body=None, cond=None):
        super(MuvNodeDoUntil, self).__init__("DO_UNTIL", pos)
        self.body = body
        self.conditional = cond

    def generate_code(self, ctx):
        ctx.scope_push()
        out = (
            "begin\n"
            "{body}\n"
            "{cond}\n"
            "until"
        ).format(
            cond=self.indent(self.conditional.generate_code(ctx)),
            body=self.indent(self.body.generate_code(ctx)),
        )
        ctx.scope_pop()
        return self.annotate(ctx, out)


class MuvNodeSwitch(MuvNode):
    def __init__(self, pos, expr=None, var=None, cases=[]):
        super(MuvNodeSwitch, self).__init__("SWITCH", pos)
        self.expr = expr
        self.var = var
        self.children = cases

    def generate_code(self, ctx):
        ctx.scope_push()
        out = (
            "begin\n"
            "{expr}\n"
            "{setexp}\n"
            "{cases}\n"
            "repeat"
        ).format(
            expr=self.indent(self.expr.generate_code(ctx)),
            setexp=self.indent(self.var.set_expr(ctx)),
            cases=self.indent(
                "\n".join(
                    case.generate_code(ctx) for case in self.children
                )
            )
        )
        ctx.scope_pop()
        return self.annotate(ctx, out)


class MuvNodeTryCatch(MuvNode):
    def __init__(self, pos, body, handler, var=None):
        super(MuvNodeTryCatch, self).__init__("TRY_CATCH", pos)
        self.body = body
        self.handler = handler
        self.settable = var

    def generate_code(self, ctx):
        ctx.scope_push()
        vexp = self.settable.set_expr(ctx) if self.settable else "pop"
        out = (
            "0 try\n"
            "{body}\n"
            "catch_detailed {vexp}\n"
            "{handler}\n"
            "endcatch"
        ).format(
            body=self.indent(self.body.generate_code(ctx)),
            handler=self.indent(self.handler.generate_code(ctx)),
            vexp=vexp,
        )
        ctx.scope_pop()
        return self.annotate(ctx, out)


class MuvNodeNamespace(MuvNode):
    def __init__(self, pos, namespace, body):
        super(MuvNodeNamespace, self).__init__("NAMESPACE", pos)
        self.namespace = namespace
        self.body = body

    def generate_code(self, ctx):
        curr_ns = ctx.curr_namespace
        ctx.set_namespace(self.namespace)
        bodyparts = [
            item.generate_code(ctx)
            for item in self.body
        ]
        out = "\n".join(x for x in bodyparts if x)
        ctx.set_namespace(curr_ns)
        return self.annotate(ctx, out)


class MuvNodeUsingNamespace(MuvNode):
    def __init__(self, pos, namespace):
        super(MuvNodeUsingNamespace, self).__init__("USING_NS", pos)
        self.namespace = namespace

    def generate_code(self, ctx):
        ctx.use_namespace(self.namespace)
        return ''


class MuvNodeExtern(MuvNode):
    def __init__(
                self, pos, extern,
                argcount=0,
                retcount=1,
                varargs=False,
                code=None
            ):
        super(MuvNodeExtern, self).__init__("NAMESPACE", pos)
        self.extern = extern
        self.argcount = argcount
        self.retcount = retcount
        self.varargs = varargs
        self.code = textwrap.dedent(code) if code else extern

    def generate_code(self, ctx):
        ctx.declare_function(
            self.extern,
            self.argcount,
            varargs=self.varargs,
            retcount=self.retcount,
            is_extern=True,
            pos=self.position,
            code=self.code,
        )
        return ''


class MuvNodeFuncDef(MuvNode):
    def __init__(
                self, pos,
                funcname=None,
                args=[],
                body=None,
                varargs=False,
                public=False,
            ):
        super(MuvNodeFuncDef, self).__init__("FUNC_DEF", pos)
        self.funcname = funcname
        self.args = args
        self.public = public
        self.varargs = varargs
        if body and type(body) in [MuvNodeStatements, MuvNodeExprList]:
            last = body.children[-1]
            if not isinstance(last, MuvNodeCommand) or last.command != "exit":
                body.append(
                    MuvNodeCommand(
                        last.position, "exit",
                        MuvNodeInteger(last.position, 0)
                    )
                )
        last = body
        self.lastpos = None
        while last and type(last) in [MuvNodeStatements, MuvNodeExprList]:
            self.lastpos = last.position
            last = last.children[-1]
        if isinstance(last, MuvNode):
            self.lastpos = last.position
        if isinstance(last, MuvNodeCommand):
            last.command = ''
            self.lastpos = last.position
        self.body = body

    def generate_code(self, ctx):
        ctx.scope_push()
        actual = ctx.declare_function(
            self.funcname,
            len(self.args),
            varargs=self.varargs,
            pos=self.position,
            is_public=self.public,
        )
        args = "".join("%s " % ctx.declare_variable(v) for v in self.args)
        fmt = ": {funcname}[ {args}-- ret ]\n{body}\n{end}"
        if self.public:
            fmt += "\npublic {funcname}\n$libdef {funcname}"
        out = fmt.format(
            funcname=actual,
            args=args,
            body=self.indent(self.body.generate_code(ctx)),
            end=self.annotate(ctx, ';', pos=self.lastpos, lineoff=1),
        )
        out = self.annotate(ctx, out)
        ctx.scope_pop()
        ctx.reset_function_local_data()
        return out


class MuvNodeFuncAddr(MuvNode):
    def __init__(self, pos, funcname=None, args=[]):
        super(MuvNodeFuncAddr, self).__init__("FUNC_ADDR", pos)
        self.funcname = funcname

    def generate_code(self, ctx):
        finfo = ctx.lookup_function(self.funcname)
        if not finfo:
            raise MuvError(
                "function '{funcname}' not found.".format(
                    funcname=self.funcname,
                ),
                position=self.position,
            )
        if finfo.is_extern:
            raise MuvError(
                "Cannot get address of extern '{funcname}'.".format(
                    funcname=self.funcname,
                ),
                position=self.position,
            )
        out = "'{func}".format(func=finfo.code)
        return self.annotate(ctx, out)


class MuvNodeIndex(MuvNode):
    def __init__(self, pos, expr=None, idx=None):
        super(MuvNodeIndex, self).__init__("INDEX", pos)
        self.expr = expr
        self.idx = idx

    def generate_code(self, ctx):
        out = "{expr} {idx} []".format(
            expr=self.expr.generate_code(ctx),
            idx=self.idx.generate_code(ctx),
        )
        return self.annotate(ctx, out)


class MuvNodeAddrCall(MuvNode):
    def __init__(self, pos, expr=None, args=[]):
        super(MuvNodeAddrCall, self).__init__("ADDR_CALL", pos)
        self.expr = expr
        self.args = args

    def generate_code(self, ctx):
        args = [x.generate_code(ctx) for x in self.args]
        out = "{args}{expr} execute".format(
            args="".join("%s " % x for x in args if x),
            expr=self.expr.generate_code(ctx),
        )
        return self.annotate(ctx, out)


class MuvNodeFuncCall(MuvNode):
    def __init__(self, pos, funcname=None, args=[]):
        super(MuvNodeFuncCall, self).__init__("FUNC_CALL", pos)
        self.funcname = funcname
        self.args = args

    def generate_code(self, ctx):
        finfo = ctx.lookup_function(self.funcname)
        if not finfo:
            raise MuvError(
                "function '{funcname}' not found.".format(
                    funcname=self.funcname,
                ),
                position=self.position,
            )
        expected = finfo.argcount
        foundcnt = len(self.args)
        ftype = "extern" if finfo.is_extern else "function"
        if finfo.varargs:
            expected -= 1
            if foundcnt < expected:
                raise MuvError(
                    "{ftype} '{funcname}' expected at least {exp} "
                    "argument(s), but found only {found}.".format(
                        funcname=self.funcname,
                        exp=expected,
                        found=foundcnt,
                        ftype=ftype
                    ),
                    position=self.position,
                )
        elif foundcnt != expected:
            raise MuvError(
                "{ftype} '{funcname}' expected {expected} argument(s), "
                "but found {found}.".format(
                    funcname=self.funcname,
                    expected=expected,
                    found=foundcnt,
                    ftype=ftype
                ),
                position=self.position,
            )
        args = []
        for arg in self.args[:expected]:
            args.append(arg.generate_code(ctx))
        if finfo.varargs:
            args.append(
                "{ %s}list" % "".join(
                    "%s " % x.generate_code(ctx)
                    for x in self.args[expected:]
                )
            )
        out = "{args}{cmd}".format(
            args="".join("%s " % x for x in args),
            cmd=finfo.code,
        )
        if finfo.retcount == 0:
            out += " 0"
        elif finfo.retcount > 1:
            out = "{ %s }list" % out
        return self.annotate(ctx, out)


class MuvNodeSourceFile(MuvNode):
    def __init__(self, pos, *children):
        super(MuvNodeSourceFile, self).__init__("SOURCEFILE", pos)
        self.children = list(children)

    def generate_code(self, ctx):
        children = [
            child.generate_code(ctx)
            for child in self.children
        ]
        out = "\n".join(x for x in children if x)
        out = self.annotate(ctx, out)
        out = self.strip_extra_annotations(out)
        return out


class MuvNodeProgram(MuvNode):
    def __init__(self, pos, code):
        super(MuvNodeProgram, self).__init__("PROGRAM", pos)
        self.code = code

    def generate_code(self, ctx):
        code = self.code.generate_code(ctx)
        inits = ctx.init_statements
        inits.append(ctx.last_function)
        if ctx.last_func_pos:
            self.position = ctx.last_func_pos
        init_code = ": __start\n{inits}\n;".format(
            inits=self.indent("\n".join(inits))
        )
        init_code = self.annotate(ctx, init_code, force=True)
        out = "{code}\n{inits}".format(
            code=code,
            inits=init_code,
        )
        replacements = []
        if ctx.optimization_level >= 1:
            replacements += [
                (r' swap swap\b', r''),
                (r' 0 pop\b', r''),
                (r'\bswap \+', r'+'),
                (r'\bswap \*', r'*'),
                (r'\bswap bitor\b', r'bitor'),
                (r'\bswap bitxor\b', r'bitxor'),
                (r'\bswap bitand\b', r'bitand'),
                (r'(\w+) @ \1 (\+\+|--) pop\b', r'\1 \2'),
            ]
        for pat, repl in replacements:
            out = re.sub(pat, repl, out)
        return out


# vim: set ts=4 sw=4 et ai hlsearch nowrap :
